@@ -71,7 +71,7 @@ def convert_unit_ltx(string: str, negative: bool = False) -> str:
             if i == "(":
                 result += '_{'
                 continue
-            elif i == ")":
+            if i == ")":
                 result += '}'
                 continue
         result += (i)
@@ -85,18 +85,14 @@ def convert_unit_ltx(string: str, negative: bool = False) -> str:
     return result
 
 
-def convert_loadingstr(string: str) -> str:
-    pass
-
-
-def _is_none(s):
+def _is_none(s: str):
     """Check if a value is a text None."""
     if s == 'None':
         return True
     return False
 
 
-def _is_float(s):
+def _is_float(s: str):
     """Check if a value is a float."""
     try:
         float(s)
@@ -105,7 +101,7 @@ def _is_float(s):
         return False
 
 
-def _is_bool(s):
+def _is_bool(s: str):
     """Check a value is a text bool."""
     if s.lower() in ['true', 'false']:
         return True
@@ -113,17 +109,16 @@ def _is_bool(s):
         return False
 
 
-def _from_bool(s):
+def _from_bool(s: str):
     """Convert a boolean into a string."""
     if s.lower() == 'true':
         return True
-    elif s.lower() == 'false':
+    if s.lower() == 'false':
         return False
-    else:
-        raise ValueError('String cannot be converted to bool')
+    raise ValueError('String cannot be converted to bool')
 
 
-def _is_list(s):
+def _is_list(s: str):
     """Check a value is a simple list."""
     if s.startswith('[') and s.endswith(']'):
         return True
@@ -131,7 +126,7 @@ def _is_list(s):
         return False
 
 
-def _from_list(s):
+def _from_list(s: str):
     """Convert a value into a string list."""
     # note that the function will fail if the list has other spaces
     return ast.literal_eval(s.replace(' ', ","))
